import Ice

import typing as ty

from armarx_core import ice_manager
from armarx_core.reconnecting_proxy_client import ReconnectingProxyClient

from armarx_memory.aron.aron_ice_types import AronIceTypes

from armarx_skills.core import execution_status as es
from armarx_skills.core.execution_status import ExecutionStatus

from armarx_skills.provider import abort_skill_result as asr
from armarx_skills.provider import dto as provider_dto

from armarx_skills.manager import dti, dto
from armarx_skills.manager import errors
from armarx_skills.manager import provider_info as pi
from armarx_skills.manager import skill_execution_id as sei
from armarx_skills.manager import skill_execution_request as ser
from armarx_skills.manager import skill_status_update as ssu
from armarx_skills.manager.skill_id import SkillID, SkillIdConv
from armarx_skills.manager import skill_description as sd


class SkillManager:
    DEFAULT_ICE_OBJECT_NAME = "SkillMemory"

    def __init__(
            self,
            proxy: dti.SkillManagerInterfacePrx = None,
    ):
        self.proxy = proxy

        self.skill_id_conv = SkillIdConv()
        self.provider_info_conv = pi.ProviderInfoConv()
        self.skill_execution_id_conv = sei.SkillExecutionIDConv()
        self.skill_execution_request_conv = ser.SkillExecutionRequestConv()
        self.skill_status_update_conv = ssu.SkillStatusUpdateConv()
        self.skill_description_conv = sd.SkillDescriptionConv()
        self.skill_description_map_conv = sd.SkillDescriptionMapConv()
        self.abort_skill_result_conv = asr.AbortSkillResultConv()

    @classmethod
    def wait_for_manager(
            cls,
            name=DEFAULT_ICE_OBJECT_NAME,
    ) -> "SkillManager":
        return cls(proxy=ice_manager.wait_for_proxy(dti.SkillManagerInterfacePrx, name))

    @classmethod
    def get_manager(
            cls,
            name=DEFAULT_ICE_OBJECT_NAME,
    ) -> ty.Optional["SkillManager"]:
        try:
            return cls(proxy=ice_manager.get_proxy(dti.SkillManagerInterfacePrx, name))
        except Ice.ConnectionRefusedException:
            return None

    def add_provider(
            self,
            provider_info: pi.ProviderInfo,
    ) -> None:
        provider_info_ice = self.provider_info_conv.to_ice(provider_info)
        return self.add_provider_ice(provider_info_ice)

    def add_provider_ice(
            self,
            provider_info: dto.ProviderInfo,
    ) -> None:
        return self.proxy.addProvider(provider_info)

    def remove_provider(
            self,
            provider_info: pi.ProviderInfo,
    ) -> None:
        provider_info_ice = self.provider_info_conv.to_ice(provider_info)
        return self.remove_provider_ice(provider_info_ice)

    def remove_provider_ice(
            self,
            provider_info: dto.ProviderInfo,
    ) -> None:
        return self.proxy.removeProvider(providerInfo=provider_info)

    def get_skill_description(
            self,
            skill_id: SkillID,
    ) -> ty.Optional[sd.SkillDescription]:
        skill_id_ice = self.skill_id_conv.to_ice(skill_id)
        description_ice = self.get_skill_description_ice(skill_id=skill_id_ice)

        if description_ice is None:
            return None
        else:
            description = self.skill_description_conv.from_ice(description_ice)
            return description

    def get_skill_description_ice(
            self,
            skill_id: dto.SkillID,
    ) -> ty.Optional[dto.SkillDescription]:
        return self._ice_unset_to_none(self.proxy.getSkillDescription(skillId=skill_id))

    def get_skill_descriptions(
            self,
    ) -> sd.SkillDescriptionMap:
        descriptions_ice = self.get_skill_descriptions_ice()
        descriptions = self.skill_description_map_conv.from_ice(descriptions_ice)
        return descriptions

    def get_skill_descriptions_ice(
            self,
    ) -> dto.SkillDescriptionMap:
        return self.proxy.getSkillDescriptions()

    def get_skill_execution_status(
            self,
            execution_id: sei.SkillExecutionID,
    ) -> ty.Optional[ssu.SkillStatusUpdate]:
        execution_id_ice = self.skill_execution_id_conv.to_ice(execution_id)
        status_ice = self.get_skill_execution_status_ice(execution_id_ice)

        if status_ice is None:
            return None
        else:
            status = self.skill_status_update_conv.from_ice(status_ice)
            return status

    def get_skill_execution_status_ice(
            self,
            execution_id: dto.SkillExecutionID,
    ) -> ty.Optional[dto.SkillStatusUpdate]:
        return self._ice_unset_to_none(self.proxy.getSkillExecutionStatus(execution_id))

    def get_skill_execution_statuses(
            self,
    ) -> ssu.SkillStatusUpdateMap:
        updates_ice = self.get_skill_execution_statuses_ice()
        updates = {
            self.skill_execution_id_conv.from_ice(k): self.skill_status_update_conv.from_ice(v)
            for k, v in updates_ice.items()
        }
        return updates

    def get_skill_execution_statuses_ice(
            self,
    ) -> dto.SkillStatusUpdateMap:
        return self.proxy.getSkillExecutionStatuses()

    def execute_skill(
            self,
            request: ser.SkillExecutionRequest,
            asynchronous=False,
            log=None,
            dry=False,
    ) -> ty.Optional[ssu.SkillStatusUpdate]:
        # Return is optional because returning None when dry == True.

        if log is not None:
            msg = "[DRY RUN] " if dry else ""
            msg += f"Executing skill {request.skill_id} with parameters: "
            for k, v in request.parameters.items():
                msg += f"\n- '{k}': {v}"
            log.info(msg)

        request_ice = self.skill_execution_request_conv.to_ice(request)

        if dry:
            return None

        if not asynchronous:
            update_ice = self.execute_skill_ice(request=request_ice)
            update: ssu.SkillStatusUpdate = self.skill_status_update_conv.from_ice(update_ice)
        else:
            skill_execution_id = self.execute_skill_async_ice(request=request_ice)
            update: ssu.SkillStatusUpdate = ssu.SkillStatusUpdate(sei.SkillExecutionIDConv().from_ice(skill_execution_id), {}, None,
                                                                  ExecutionStatus.Preparing, {})

        if log is not None:
            log.info(f"Skill {request.skill_id} terminated with status '{update.status.name}'.")

        return update

    def execute_skill_based_on_default_parameters(
            self,
            executor_name: str,
            skill_id: SkillID,
            parameters_function: ty.Callable[[ty.Dict], None],
            asynchronous=False,
            log=None,
            dry=False

    ) -> ssu.SkillStatusUpdate:
        """
        Execute a skill based on its default parameters.

        parameters_function is a function that edits the default parameters
        before executing the skill.
        """
        descriptions = self.get_skill_descriptions()
        try:
            description = descriptions[skill_id]
        except KeyError as e:
            raise errors.SkillNotFoundError(skill_id=skill_id)

        parameters = description.root_profile_defaults

        parameters_function(parameters)

        return self.execute_skill(ser.SkillExecutionRequest(
            executor_name=executor_name,
            skill_id=skill_id,
            parameters=parameters),
            asynchronous=asynchronous,
            log=log,
            dry=dry)

    def execute_skill_with_default_parameters(
            self,
            executor_name: str,
            skill_id: SkillID,
            log=None,
            dry=False,
    ) -> ssu.SkillStatusUpdate:
        """
        Execute a skill with its default parameters.
        """
        return self.execute_skill_based_on_default_parameters(
            executor_name=executor_name,
            skill_id=skill_id,
            parameters_function=lambda parameters: None,
            log=log,
            dry=dry,
        )

    def execute_skill_ice(
            self,
            request: dto.SkillExecutionRequest,
    ) -> dto.SkillStatusUpdate:
        return self.proxy.executeSkill(skillExecutionRequest=request)

    def execute_skill_async_ice(
            self,
            request: dto.SkillExecutionRequest,
    ) -> dto.SkillExecutionID:
        return self.proxy.executeSkillAsync(skillExecutionRequest=request)

    def update_skill_parameters_ice(
            self,
            execution_id: dto.SkillExecutionID,
            parameters: AronIceTypes.Dict,
    ) -> dto.SkillExecutionID:
        return self.proxy.updateSkillParameters(
            executionId=execution_id,
            parameters=parameters,
        )

    def abort_skill(
            self,
            skill_execution_id: sei.SkillExecutionID,
            blocking=True,
    ) -> asr.AbortSkillResult:
        skill_execution_id_ice = self.skill_execution_id_conv.to_ice(skill_execution_id)

        abort_result_ice = self.abort_skill_ice(
            execution_id=skill_execution_id_ice,
            blocking=blocking,
        )

        abort_result = self.abort_skill_result_conv.from_ice(abort_result_ice)
        return abort_result

    def abort_skill_async(
            self,
            skill_execution_id: sei.SkillExecutionID,
    ) -> asr.AbortSkillResult:
        skill_execution_id_ice = self.skill_execution_id_conv.to_ice(skill_execution_id)

        abort_result_ice = self.abort_skill_async_ice(
            execution_id=skill_execution_id_ice,
        )

        abort_result = self.abort_skill_result_conv.from_ice(abort_result_ice)
        return abort_result

    def abort_skill_ice(
            self,
            execution_id: dto.SkillExecutionID,
            blocking=True,
    ) -> provider_dto.AbortSkillResult:
        if blocking:
            return self.proxy.abortSkill(executionId=execution_id)
        else:
            return self.abort_skill_async_ice(execution_id=execution_id)

    def abort_skill_async_ice(
            self,
            execution_id: dto.SkillExecutionID,
    ) -> provider_dto.AbortSkillResult:
        return self.proxy.abortSkillAsync(executionId=execution_id)


    def abort_all_non_terminated_skills(
            self,
            abort_statuses=(
                    es.ExecutionStatus.Constructing,
                    es.ExecutionStatus.Initializing,
                    es.ExecutionStatus.Preparing,
                    es.ExecutionStatus.Running,
            ),
            blocking=True,
            log=None,
    ) -> ty.Dict[sei.SkillExecutionID, asr.AbortSkillResult]:
        # temporary fix until RobotAPI is merged
        """execution_ids_to_abort = list()

        execution_status_map = self.get_skill_execution_statuses()

        for execution_id, execution_status in execution_status_map.items():
            if execution_status.status in abort_statuses:
                log.info(f"{execution_id} RUNNING")
                execution_ids_to_abort.append(execution_id)
            else:
                log.info(f"{execution_id} not running")

        results = dict()
        for execution_id in execution_ids_to_abort:
            if log is not None:
                log.info(f"Aborting skill execution {execution_id} ...")
            result = self.abort_skill(skill_execution_id=execution_id, blocking=blocking)
            results[execution_id] = result

        return results"""

        return self.proxy.abortAllSkillsAsync()

    @staticmethod
    def _ice_unset_to_none(optional):
        return None if optional == Ice.Unset else optional


class ReconnectingSkillManager(ReconnectingProxyClient):

    def __init__(
            self,
            name=SkillManager.DEFAULT_ICE_OBJECT_NAME,
    ):
        super().__init__()
        self.name = name

    def get(self, wait=True, log=None, logger=None) -> ty.Optional[SkillManager]:
        if logger is not None:
            logger.info("The parameter 'logger' is deprecated. Use 'log' instead.")
            if log is None:
                log = logger
            else:
                raise ValueError("The parameter 'logger' is deprecated. Use 'log' instead.")

        return self._get(wait=wait, log=log)

    def _get_client(self, log=None) -> ty.Optional[SkillManager]:
        if log is not None:
            log.info(f"Waiting for skill manager '{self.name}' ...")

        client = SkillManager.get_manager(name=self.name)

        if log is not None:
            log.info(f"Connected to skill manager '{self.name}'.")

        return client

    def _wait_for_client(self, log=None) -> SkillManager:
        if log is not None:
            log.info(f"Getting skill manager '{self.name}' ...")

        client = SkillManager.wait_for_manager(name=self.name)

        if log is not None:
            if client is None:
                log.info(f"Failed to connect to skill manager '{self.name}'.")
            else:
                log.info(f"Connected to skill manager '{self.name}'.")

        return client


if __name__ == '__main__':
    def test_main():
        manager = SkillManager.wait_for_manager()
        manager.get_skill_execution_statuses()


    test_main()
